#define BOOST_ALL_NO_LIB
#include <windows.h>
#include "avisynth.h"
#include <deque>
#include <algorithm>

#include <iostream>

#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/move/move.hpp>
#include <boost/container/vector.hpp>
#include <boost/date_time/posix_time/posix_time_types.hpp>

#include "lua.hpp"

using boost::thread;
using boost::recursive_mutex;
using std::deque;
using boost::condition_variable_any;
using boost::unique_lock;


#include "getnextframe.h"

struct EmptyType {};

template<bool condition, class T1, class T2>
struct ConditionType;

template<class T1, class T2>
struct ConditionType<true, T1, T2>
{
    typedef T1 Type;
};

template<class T1, class T2>
struct ConditionType<false, T1, T2>
{
    typedef T2 Type;
};

typedef boost::container::vector<char> DataBuffer;

template<bool useCopy>
class BufferedFrameImpl
{
public:
    int nFrame; //֡
    int nLifeTimeLeft; //ʼֵڻ֡ÿһ֡ͼһС0ʱ
    typename ConditionType<useCopy, DataBuffer, PVideoFrame>::Type frame; //֡
    typename ConditionType<useCopy, EmptyType, IScriptEnvironment*>::Type env;
};

template<bool useCopy>
class BufferedFrame
{
public:
    BOOST_COPYABLE_AND_MOVABLE(BufferedFrame<useCopy>)
    boost::shared_ptr<BufferedFrameImpl<useCopy> > impl;
    
    const BufferedFrame<useCopy>& operator = (BOOST_COPY_ASSIGN_REF(BufferedFrame<useCopy>) bf)
    {
        if(this != &bf)
        {
            impl->nFrame = bf->nFrame;
            impl->nLifeTimeLeft = bf->nLifeTimeLeft;
            impl->frame = bf->frame;
            impl->env = bf->env;
        }
        return *this;
    }
    
    const BufferedFrame<useCopy>& operator = (BOOST_RV_REF(BufferedFrame<useCopy>) bf)
    {
        if(this != &bf)
        {
            bf.impl.swap(impl);
        }
        return *this;
    }
    
    BufferedFrame(const BufferedFrame<useCopy>& a)
    {
        impl.reset(new BufferedFrameImpl<useCopy>);
        *this = a;
    }
    
    BufferedFrame(BOOST_RV_REF(BufferedFrame<useCopy>) bf)
    {
        *this = boost::move(bf);
    }
    
    BufferedFrame()
    {
        impl.reset(new BufferedFrameImpl<useCopy>);
    }
    
    static inline int GetFrameDataSize(PVideoFrame& frame)
    {
        return frame->GetFrameBuffer()->GetDataSize() - frame->GetOffset();
    }
    
    void LoadFrame(IScriptEnvironment* env, PVideoFrame f);
    
    PVideoFrame GetFrame(IScriptEnvironment* env, const VideoInfo& vi);
    
    BufferedFrameImpl<useCopy>* operator ->()
    {
        return impl.get();
    }
};

template<bool useCache, bool useCopy>
struct SoraThreadHelper;

template<bool useCache, bool useCopy>
class SoraThread : public GenericVideoFilter
{
        friend class SoraThreadHelper<useCache, useCopy>;
    private:
        //֡
        int nFrameCount;
        
        //棬֡ݴ
        int nCacheCount;
        typename ConditionType<useCache, deque<BufferedFrame<useCopy> >, EmptyType>::Type frameCache;
        typename ConditionType<useCache, recursive_mutex, EmptyType>::Type frameCacheMutex;
        
        //壬û֡ݴԺƶ
        int nBufferCount;
        deque<BufferedFrame<useCopy> > frameBuffer;
        recursive_mutex frameBufferMutex;
        
        //ʱ֪ͨ
        condition_variable_any frameBufferChangedEvent;
        
        //߳
        thread processThread;
        
        //һ֡ʲô
        int nNextFrameToRead;
        recursive_mutex nextFrameToReadMutex;
        
        //һ֡ķ
        IGetNextFrame* nextFrameGetter;
        
        //֪ͨ߳
        recursive_mutex startEventMutex;
        condition_variable_any startEvent;
        
        IScriptEnvironment* firstEnv; //һͽenv
        
    public:
        SoraThread(IScriptEnvironment* env, PClip _child, int bc, int cc, const char* luascript)
            : GenericVideoFilter(_child)
            , nCacheCount(cc)
            , nBufferCount(bc)
        {
            nFrameCount = vi.num_frames;
            nNextFrameToRead = -1;
            if(luascript == 0)
                nextFrameGetter = new GetNextFrameDefault(nFrameCount);
            else
            {
                try
                {
                    nextFrameGetter = new GetNextFrameLua(luascript, nFrameCount);
                }
                catch(GetNextFrameLua::ScriptError)
                {
                    delete nextFrameGetter;
                    env->ThrowError("SoraThread:  Incorrect lua script.");
                }
            }
        }
        
        __stdcall ~SoraThread()
        {
            if( processThread.get_id() != boost::thread::id() )
            {
                processThread.interrupt();
                processThread.join();
            }
            delete nextFrameGetter;
        }
        
        PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env)
        {
            PVideoFrame frame;
            if( processThread.get_id() == boost::thread::id() )
            {
                firstEnv = env;
                processThread = thread(boost::bind(&SoraThread::threadFunction, this));
                nNextFrameToRead = n;
                WaitForFrame(env, n, frame);
                
                return frame;
            }
            if( SearchCache(env, n, frame) ) //ڻҵҪ֡
            {
                ;
            }
            else if( SearchBuffer(env, n, frame) ) //ڻҵҪ֡
            {
                ;
            }
            else //ûҵ趨һ֡Ժȴ
            {
                {
                    unique_lock<recursive_mutex> lg(nextFrameToReadMutex);
                    nNextFrameToRead = n;
                }
                frameBufferChangedEvent.notify_all();
                WaitForFrame(env, n, frame);
            }
            
            return frame;
        }
        
    private:
        bool SearchCache(IScriptEnvironment* env, int nFrame, PVideoFrame& frame) //һ棬ҵԺµڵһ
        {
            return SoraThreadHelper<useCache, useCopy>::SearchCache(*this, env, nFrame, frame);
        }
        
        bool SearchBuffer(IScriptEnvironment* env, int nFrame, PVideoFrame& frame) //һ
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            deque<BufferedFrame<useCopy> >::iterator i = frameBuffer.begin();
            
            while(i != frameBuffer.end())
            {
                if((*i)->nFrame == nFrame)
                {
                    frame = i->GetFrame(env, vi);
                    CacheFrame(nFrame, boost::move(*i), env);
                    i = frameBuffer.erase(i);
                    lg.unlock();
                    frameBufferChangedEvent.notify_all();
                    return true;
                }
                ++i;
            }
            return false;
        }
        
        void CacheFrame(int nFrame, BOOST_RV_REF(BufferedFrame<useCopy>) frame, IScriptEnvironment* env)
        {
            SoraThreadHelper<useCache, useCopy>::CacheFrame(*this, nFrame, boost::move(frame), env);
        }
        
        void BufferFrame(int nFrame, PVideoFrame frame, IScriptEnvironment* env)
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            while(frameBuffer.size() >= nBufferCount) //˻ȴ
            {
                frameBufferChangedEvent.timed_wait(lg, boost::posix_time::time_duration(0,0,2));
            }
            
            ReduceBufferedFrameLifeTime();
            
            BufferedFrame<useCopy> bf;
            bf.LoadFrame(env, frame);
            bf->nFrame = nFrame;
            bf->nLifeTimeLeft = nBufferCount;
            frameBuffer.push_back(boost::move(bf));
            
            lg.unlock();
            frameBufferChangedEvent.notify_all();
        }
        
        void ReduceBufferedFrameLifeTime()
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            deque<BufferedFrame<useCopy> >::iterator i = frameBuffer.begin();
            while(i != frameBuffer.end())
            {
                -- (*i)->nLifeTimeLeft;
                if( (*i)->nLifeTimeLeft <= 0 )
                    i = frameBuffer.erase(i);
                else
                    ++i;
            }
            
            lg.unlock();
            frameBufferChangedEvent.notify_all();
        }
        
        void WaitForFrame(IScriptEnvironment* env, int nFrame, PVideoFrame& frame)
        {
            unique_lock<recursive_mutex> lg(frameBufferMutex);
            
            while(SearchBuffer(env, nFrame, frame) == false)
            {
                while(frameBuffer.size() >= nBufferCount)
                    ReduceBufferedFrameLifeTime();
                
                startEvent.notify_all();
                frameBufferChangedEvent.timed_wait(lg, boost::posix_time::time_duration(0,0,2));
            }
        }
        
        void threadFunction()
        {
            try
            {
                int nLastReadFrame = -1;
                int nFrameToRead = -1;
                for(;;)
                {
                    boost::this_thread::interruption_point();
                    {
                        unique_lock<recursive_mutex> lg(nextFrameToReadMutex);
                        nFrameToRead = nNextFrameToRead;
                        nNextFrameToRead = GetNextFrameToRead(nNextFrameToRead, firstEnv);
                    }
                    {
                        unique_lock<recursive_mutex> lg(startEventMutex);
                        while(nFrameToRead == -1)
                        {
                            startEvent.timed_wait(lg, boost::posix_time::time_duration(0,0,2));
                            {
                                unique_lock<recursive_mutex> lg2(nextFrameToReadMutex);
                                nFrameToRead = nNextFrameToRead;
                            }
                        }
                    }
                    
                    if(nLastReadFrame != nFrameToRead) //Ϊ˽ ߳2ڶ߳1󡪡߳2ꡪ߳1ߡ߳2õ󡪡ֻûСֶ
                    {
                        PVideoFrame vf = child->GetFrame(nFrameToRead, firstEnv);
                        
                        nLastReadFrame = nFrameToRead;
                        BufferFrame(nFrameToRead, vf, firstEnv);
                    }
                }
            }
            catch(boost::thread_interrupted)
            {
                ;
            }
        }
        
        int GetNextFrameToRead(int nCurrentFrame, IScriptEnvironment* env)
        {
            try
            {
                return nextFrameGetter->GetNextFrame(nCurrentFrame);
            }
            catch(GetNextFrameLua::ScriptError)
            {
                env->ThrowError("SoraThread:  lua script runs into error.");
                return -2;
            }
        }
};

void BufferedFrame<true>::LoadFrame(IScriptEnvironment* env, PVideoFrame f)
{
    int len = GetFrameDataSize(f);
    const char* srcData = (const char*) f->GetReadPtr();
    impl->frame = DataBuffer(srcData, srcData+len);
}

PVideoFrame BufferedFrame<true>::GetFrame(IScriptEnvironment* env, const VideoInfo& vi)
{
    PVideoFrame f = env->NewVideoFrame(vi);
    if(!f->IsWritable())
        env->MakeWritable(&f);
    
    char* dstData = (char*) f->GetWritePtr();
    int len = min( impl->frame.size(), GetFrameDataSize(f) );
    copy(impl->frame.begin(), impl->frame.begin()+len, dstData);
    return f;
}

void BufferedFrame<false>::LoadFrame(IScriptEnvironment* env, PVideoFrame f)
{
    impl->frame = f;
    impl->env = env;
}

PVideoFrame BufferedFrame<false>::GetFrame(IScriptEnvironment* env, const VideoInfo& vi)
{
    if(impl->env == env)
    {
        return impl->frame;
    }
    else
    {
        PVideoFrame f = env->NewVideoFrame(vi);
        if(!f->IsWritable())
            env->MakeWritable(&f);
        
        int len = min( GetFrameDataSize(f), GetFrameDataSize(impl->frame) );
        const char* srcData = (const char*) impl->frame->GetReadPtr();
        char* dstData = (char*) f->GetWritePtr();
        std::copy(srcData, srcData+len, dstData);
        return f;
    }
}

template<bool useCopy>
struct SoraThreadHelper<true, useCopy>
{
    static void CacheFrame(SoraThread<true, useCopy>& obj, int& nFrame, BOOST_RV_REF(BufferedFrame<useCopy>) frame, IScriptEnvironment* env)
    {
        unique_lock<recursive_mutex> lg(obj.frameCacheMutex);
        
        while(obj.frameCache.size() >= obj.nCacheCount)
            obj.frameCache.pop_back();
        
        obj.frameCache.push_front(boost::move(frame));
    }
    
    static bool SearchCache(SoraThread<true, useCopy>& obj, IScriptEnvironment* env, int& nFrame, PVideoFrame& frame)
    {
        unique_lock<recursive_mutex> lg(obj.frameCacheMutex);
        
        deque<BufferedFrame<useCopy> >::iterator i = obj.frameCache.begin();
        
        while(i != obj.frameCache.end())
        {
            if((*i)->nFrame == nFrame)
            {
                frame = i->GetFrame(env, obj.vi);
                std::swap(*i, *obj.frameCache.begin());
                return true;
            }
            ++i;
        }
        return false;
    }
};

template<bool useCopy>
struct SoraThreadHelper<false, useCopy>
{
    static inline void CacheFrame(SoraThread<false, useCopy>& obj, int& nFrame, BOOST_RV_REF(BufferedFrame<useCopy>) frame, IScriptEnvironment* env)
    {
    }

    static inline bool SearchCache(SoraThread<false, useCopy>& obj, IScriptEnvironment* env, int& nFrame, PVideoFrame& frame)
    {
        return false;
    }
};

AVSValue __cdecl Create_SoraThread(AVSValue args, void* user_data, IScriptEnvironment* env)
{
    if(args[0].IsClip())
    {
        int bufferCount = args[1].IsInt() ? args[1].AsInt() : 5;
        int cacheCount = args[2].IsInt() ? args[2].AsInt() : 0;
        const char* preFetch = args[3].IsString() ? args[3].AsString() : 0;
        bool useCopy = args[4].IsBool() ? args[4].AsBool() : false;
        if(cacheCount == 0)
        {
            if(useCopy)
                return AVSValue(new SoraThread<false, true>(env, args[0].AsClip(), bufferCount, cacheCount, preFetch));
            else
                return AVSValue(new SoraThread<false, false>(env, args[0].AsClip(), bufferCount, cacheCount, preFetch));
        }
        else
        {
            if(useCopy)
                return AVSValue(new SoraThread<true, true>(env, args[0].AsClip(), bufferCount, cacheCount, preFetch));
            else
                return AVSValue(new SoraThread<true, false>(env, args[0].AsClip(), bufferCount, cacheCount, preFetch));
        }
    }
    else
    {
        env->ThrowError("SoraThread:  No available clip!");
        return AVSValue(0);
    }
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("SoraThread", "c[Buffer]i[Cache]i[Prefetch]s[CopyData]b", Create_SoraThread, 0);
    return 0;
}
